;; Generated package description from gited.el  -*- no-byte-compile: t -*-
(define-package "gited" "0.6.0.0.20221221.75709" "Operate on Git branches like dired" '((emacs "24.4") (cl-lib "0.5")) :commit "506686776e541f22b84ef64fef13b555e6b3bf13" :url "https://elpa.gnu.org/packages/gited.html" :authors '(("Tino Calancha" . "tino.calancha@gmail.com")) :maintainer '("Tino Calancha" . "tino.calancha@gmail.com") :keywords '("git" "vc" "convenience"))
